<?php

namespace KnitPay\Extensions\LearnDash;

use Pronamic\WordPress\Money\Money;
use Pronamic\WordPress\Money\TaxedMoney;
use Pronamic\WordPress\Money\Parser as MoneyParser;
use Pronamic\WordPress\Pay\Payments\PaymentLines;
use Pronamic\WordPress\Pay\Payments\PaymentData as Pay_PaymentData;
use Pronamic\WordPress\Pay\Payments\Item;
use Pronamic\WordPress\Pay\Payments\Items;
use LearnDash_Settings_Section;
use LearnDash_User_Status_Widget;
use Pronamic\WordPress\Pay\Payments\PaymentDataInterface as Pronamic_Pay_PaymentDataInterface;

/**
 * Title: Learn Dash LMS Payment Data
 * Description:
 * Copyright: 2020 Knit Pay
 * Company: Knit Pay
 *
 * @author  knitpay
 * @since   2.7.0
 */
class PaymentData extends Pay_PaymentData {

	/**
	 * Gateway
	 *
	 * @var Gateway
	 */
	private $gateway;

	/**
	 * User data
	 *
	 * @var array
	 */
	private $user_data;

	/**
	 * Constructs and initializes an LearnPress payment data object.
	 *
	 * @param int $course_id   Course ID of Learn Dash Course
	 */
	public function __construct( $course_id, $config ) {
		$this->course_id = $course_id;
		$this->order_id  = $this->course_id;
		$this->config    = $config;

		$this->init( $course_id );

		parent::__construct();
	}

	private function init( $course_id ) {
		 $course = get_post( $course_id );

		if ( ! $course ) {
			return false;
		}

		$user_id    = get_current_user_id();
		$user_email = null;

		if ( 0 != $user_id ) {
			$this->user = get_userdata( $user_id );
		}

		if ( learndash_get_post_type_slug( 'course' ) === $course->post_type ) {
			$course_price          = learndash_get_setting( $course->ID, 'course_price' );
			$course_price_type     = learndash_get_setting( $course->ID, 'course_price_type' );
			$course_plan_id        = 'learndash-course-' . $course->ID;
			$course_interval_count = get_post_meta( $course->ID, 'course_price_billing_p3', true );
			$course_interval       = get_post_meta( $course->ID, 'course_price_billing_t3', true );
		} elseif ( learndash_get_post_type_slug( 'group' ) === $course->post_type ) {
			$course_price          = learndash_get_setting( $course->ID, 'group_price' );
			$course_price_type     = learndash_get_setting( $course->ID, 'group_price_type' );
			$course_plan_id        = 'learndash-group-' . $course->ID;
			$course_interval_count = get_post_meta( $course->ID, 'group_price_billing_p3', true );
			$course_interval       = get_post_meta( $course->ID, 'group_price_billing_t3', true );
		}

		switch ( $course_interval ) {
			case 'D':
				$course_interval = 'day';
				break;

			case 'W':
				$course_interval = 'week';
				break;

			case 'M':
				$course_interval = 'month';
				break;

			case 'Y':
				$course_interval = 'year';
				break;
		}

		$course_image = get_the_post_thumbnail_url( $course->ID, 'medium' );
		$course_name  = $course->post_title;
		$course_id    = $course->ID;

		$course_price = preg_replace( '/.*?(\d+(?:\.?\d+))/', '$1', $course_price );

		$args               = compact( 'user_id', 'course_id', 'course_image', 'course_name', 'course_plan_id', 'course_price', 'course_price_type', 'course_interval', 'course_interval_count' );
		$this->payment_info = $args;

	}

	/**
	 * Get source indicator
	 *
	 * @see Pronamic_Pay_PaymentDataInterface::get_source()
	 * @return string
	 */
	public function get_source() {
		return 'learndash';
	}

	/**
	 * Get source ID.
	 *
	 * @return string
	 */
	public function get_source_id() {
		return $this->order_id;
	}

	/**
	 * Get title.
	 *
	 * @return string
	 */
	public function get_title() {
		return $this->payment_info['course_name'];
	}

	/**
	 * Get description
	 *
	 * @see Pronamic_Pay_PaymentDataInterface::get_description()
	 * @return string
	 */
	public function get_description() {
		$description = $this->config['payment_description'];

		if ( '' === $description ) {
			$description = $this->get_title();
		}

		// Replacements.
		$replacements = array(
			'{course_id}'   => $this->get_order_id(),
			'{course_name}' => $this->payment_info['course_name'],
		);

		return strtr( $description, $replacements );
	}

	/**
	 * Get order ID
	 *
	 * @see Pronamic_Pay_PaymentDataInterface::get_order_id()
	 * @return string
	 */
	public function get_order_id() {
		return $this->order_id;
	}

	/**
	 * Get items
	 *
	 * @see Pronamic_Pay_PaymentDataInterface::get_items()
	 * @return Items
	 */
	/*
	  public function get_items() {

		$payment_lines = new PaymentLines();
		if ( $items = LP()->get_cart()->get_items() ) {
			foreach ( $items as $item ) {

					$line = $payment_lines->new_line();

					$line->set_id( $item['item_id'] );
					$line->set_name( get_the_title( $item['item_id'] ) );
					$line->set_quantity( $item['quantity'] );
					$line->set_discount_amount( new Money($item['data']->get_sale_price(), $this->get_currency_alphabetic_code()) );
					$line->set_unit_price( new TaxedMoney($item['data']->get_origin_price(), $this->get_currency_alphabetic_code()) );
					$line->set_total_amount( new TaxedMoney($item['data']->get_sale_price(), $this->get_currency_alphabetic_code()) );
					$line->set_product_url( get_permalink( $item['item_id'] ) );
				}
			}

			ppp($payment_lines);ddd();
	} */
	public function get_items() {

	}

	public function get_amount() {
		$amount = new TaxedMoney( $this->payment_info['course_price'], $this->get_currency_alphabetic_code() );
		return $amount;
	}

	/**
	 * Get currency
	 *
	 * @see LearnDash_User_Status_Widget::learndash_30_get_currency_symbol()
	 * @see Pronamic_Pay_PaymentDataInterface::get_currency_alphabetic_code()
	 * @return string
	 */
	public function get_currency_alphabetic_code() {

		$options          = get_option( 'sfwd_cpt_options' );
		$currency_setting = class_exists( 'LearnDash_Settings_Section' ) ? LearnDash_Settings_Section::get_section_setting( 'LearnDash_Settings_Section_PayPal', 'paypal_currency' ) : null;
		$currency         = '';
		$stripe_settings  = get_option( 'learndash_stripe_settings' );

		if ( ! function_exists( 'is_plugin_active' ) ) {
			include_once ABSPATH . 'wp-admin/includes/plugin.php';
		}
		if ( is_plugin_active( 'learndash-stripe/learndash-stripe.php' ) && ! empty( $stripe_settings ) && ! empty( $stripe_settings['currency'] ) ) {
			$currency = $stripe_settings['currency'];
		} elseif ( isset( $currency_setting ) || ! empty( $currency_setting ) ) {
			$currency = $currency_setting;
		} elseif ( isset( $options['modules'] ) && isset( $options['modules']['sfwd-courses_options'] ) && isset( $options['modules']['sfwd-courses_options']['sfwd-courses_paypal_currency'] ) ) {
			$currency = $options['modules']['sfwd-courses_options']['sfwd-courses_paypal_currency'];
		}

		return $currency;
	}

	/**
	 * Get email.
	 *
	 * @return string
	 */
	public function get_email() {
		if ( isset( $this->user->user_email ) ) {
			return $this->user->user_email;
		}
		return '';
	}

	/**
	 * Get first name.
	 *
	 * @return string
	 */
	public function get_first_name() {
		if ( isset( $this->order->billing_first_name ) ) {
			return $this->order->billing_first_name;
		}
	}

	/**
	 * Get last name.
	 *
	 * @return string
	 */
	public function get_last_name() {
		if ( isset( $this->order->billing_last_name ) ) {
			return $this->order->billing_last_name;
		}
	}

	/**
	 * Get customer name.
	 *
	 * @return string
	 */
	public function get_customer_name() {
		return $this->order->get_customer_name();
	}

	/**
	 * Get telephone number.
	 *
	 * @return null|string
	 */
	public function get_telephone_number() {
		// TODO
		return '';
	}

	/**
	 * Get normal return URL.
	 *
	 * @link https://github.com/woothemes/woocommerce/blob/v2.1.3/includes/abstracts/abstract-wc-payment-gateway.php#L52
	 * @return string
	 */
	public function get_normal_return_url() {
		ddd();
		ppp( esc_url( $this->gateway->get_return_url( $this->order ) ) );
		ddd();
		return esc_url( $this->gateway->get_return_url( $this->order ) );

		// TODO:
		return charitable_get_permalink( 'donation_receipt_page', array( 'donation_id' => $this->donation_id ) );
	}

	/**
	 * Get cancel URL.
	 *
	 * @return string
	 */
	public function get_cancel_url() {
		ddd();
		return esc_url( learn_press_is_enable_cart() ? learn_press_get_page_link( 'cart' ) : get_home_url() /* SITE_URL */ );
		$cancel_url = charitable_get_permalink( 'donation_cancel_page', array( 'donation_id' => $this->donation_id ) );

		if ( ! $cancel_url ) {
			$cancel_url = esc_url_raw(
				add_query_arg(
					array(
						'donation_id' => $this->donation_id,
						'cancel'      => true,
					),
					wp_get_referer()
				)
			);
		}

		return $cancel_url;
	}

	/**
	 * Get success URL.
	 *
	 * @return string
	 */
	public function get_success_url() {
		ddd();
		return esc_url( $this->gateway->get_return_url( $this->order ) );
	}

	/**
	 * Get error URL.
	 *
	 * @return string
	 */
	public function get_error_url() {
		ddd();
		return esc_url( learn_press_is_enable_cart() ? learn_press_get_page_link( 'cart' ) : get_home_url() /* SITE_URL */ );
		$cancel_url = charitable_get_permalink( 'donation_cancel_page', array( 'donation_id' => $this->donation_id ) );

		if ( ! $cancel_url ) {
			$cancel_url = esc_url_raw(
				add_query_arg(
					array(
						'donation_id' => $this->donation_id,
						'cancel'      => true,
					),
					wp_get_referer()
				)
			);
		}

		return $cancel_url;
	}
}
