<?php
namespace KnitPay\Gateways\Razorpay;

use Pronamic\WordPress\Pay\Core\Gateway as Core_Gateway;

use Pronamic\WordPress\Pay\Payments\Payment;
use Razorpay\Api\Api;
use Razorpay\Api\Errors;
use Pronamic\WordPress\Pay\Payments\PaymentStatus;
use Pronamic\WordPress\Pay\Subscriptions\Subscription;

/**
 * Title: Razorpay Gateway
 * Copyright: 2020 Knit Pay
 *
 * @author Knit Pay
 * @version 1.0.0
 * @since   1.7.0
 */
class Gateway extends Core_Gateway {
	const NAME = 'razorpay';

	/**
	 * Constructs and initializes an Razorpay gateway
	 *
	 * @param Config $config
	 *            Config.
	 */
	public function __construct( Config $config ) {
		parent::__construct( $config );

		$this->set_method( self::METHOD_HTTP_REDIRECT );

		// Supported features.
		$this->supports = array(
			'payment_status_request',
		);

		if ( defined( 'KNIT_PAY_RAZORPAY_SUBSCRIPTION' ) ) {
			$this->supports = wp_parse_args(
				$this->supports,
				array(
					'recurring_credit_card',
					'recurring',
				)
			);
		}

		$this->ENV = 'prod';
		if ( self::MODE_TEST === $config->mode ) {
			$this->ENV = 'test';
		}
	}

	/**
	 * Get supported payment methods
	 *
	 * @see Core_Gateway::get_supported_payment_methods()
	 */
	public function get_supported_payment_methods() {
		return array(
			PaymentMethods::BANK_TRANSFER,
			PaymentMethods::CREDIT_CARD,
			PaymentMethods::MAESTRO,
			PaymentMethods::RAZORPAY,
		);
	}

	/**
	 * Start.
	 *
	 * @see Core_Gateway::start()
	 *
	 * @param Payment $payment
	 *            Payment.
	 */
	public function start( Payment $payment ) {

		$payment_currency = $payment->get_total_amount()->get_currency()->get_alphabetic_code();

		$key_id     = $this->config->key_id;
		$key_secret = $this->config->key_secret;

		$api = new Api( $key_id, $key_secret );

		$customer = $payment->get_customer();

		// Payment method.
		$payment_method = $payment->get_method();

		// Recurring payment method.
		$subscription = $payment->get_subscription();

		$is_recurring_payment = ( $subscription && $this->supports( 'recurring' ) );

		if ( $is_recurring_payment ) {
			$this->get_recurring_payment_link( $api, $payment, $subscription, $customer, $payment_currency );
			return;
		}

		$this->get_payment_link( $api, $payment, $customer, $payment_currency );

	}

	/**
	 * Update status of the specified payment.
	 *
	 * @param Payment $payment
	 *            Payment.
	 */
	public function update_status( Payment $payment ) {
		if ( PaymentStatus::SUCCESS === $payment->get_status() ) {
			return;
		}

		if ( null === $payment->get_transaction_id() ) {
			return;
		}

		$is_recurring_payment = isset( $payment->subscription_id );

		$webhook_call = filter_has_var( INPUT_GET, 'kp_razorpay_webhook' );
		$key_id       = $this->config->key_id;
		$key_secret   = $this->config->key_secret;

		$api = new Api( $key_id, $key_secret );

		if ( $webhook_call ) {
			$this->handle_webhook( $api, $payment );
			return;
		}

		$invoice = $api->invoice->fetch( $payment->get_transaction_id() );
		if ( ! ( $is_recurring_payment || intval( $invoice->notes['knitpay_payment_id'] ) === $payment->get_id() ) ) {
			return;
		}

		if ( floatval( $invoice->amount ) !== $payment->get_total_amount()->format() * 100 ) {
			return;
		}

		// Return if payment id not found.
		if ( empty( $invoice->payment_id ) ) {
			return;
		}

		$razorpay_payment = $api->payment->fetch( $invoice->payment_id );

		$this->update_payment_status( $payment, $razorpay_payment );
	}

	private function update_payment_status( $payment, $razorpay_payment ) {
		$payment->add_note( 'Razorpay Status: ' . $razorpay_payment->status );
		$payment->set_status( Statuses::transform( $razorpay_payment->status ) );

		if ( PaymentStatus::SUCCESS === $payment->get_status() ) {
			$payment->set_transaction_id( $razorpay_payment->id );
		}
	}

	private function get_payment_link( $api, $payment, $customer, $payment_currency ) {

		// TODO: Add line items
		$link_data = array(
			'type'            => 'link',
			'amount'          => $payment->get_total_amount()->format() * 100,
			'currency'        => $payment_currency,
			'reminder_enable' => false,
			'sms_notify'      => false,
			'email_notify'    => false,
			'description'     => $payment->get_description(),
			'customer'        => array(
				'name'    => substr( trim( ( html_entity_decode( $customer->get_name(), ENT_QUOTES, 'UTF-8' ) ) ), 0, 45 ),
				'email'   => $customer->get_email(),
				'contact' => $payment->get_billing_address()->get_phone(),
			),
			'notes'           => array(
				'knitpay_payment_id' => $payment->id,
			),
			'callback_url'    => $payment->get_return_url(),
			'callback_method' => 'get',
		);

		$razorpay_invoice = $api->invoice->create(
			$link_data
		);

		// Update gateway results in payment.
		$payment->set_transaction_id( $razorpay_invoice->id );
		$payment->set_action_url( $razorpay_invoice->short_url );
	}

	private function get_recurring_payment_link( $api, Payment $payment, Subscription $subscription, $customer, $payment_currency ) {
		if ( 1 < count( $subscription->get_payments() ) ) {
			return;
		}

		$subscription_period = $payment->get_periods()[0];

		$subscription_phase = $subscription_period->get_phase();

		switch ( substr( $subscription_phase->get_interval()->get_specification(), -1, 1 ) ) {
			case 'D':
				$period = 'daily';
				break;
			case 'W':
				$period = 'weekly';
				break;
			case 'M':
				$period = 'monthly';
				break;
			case 'Y':
				$period = 'yearly';
				break;
			default:
				return;
		}
		$plan_data     = array(
			'period'   => $period,
			'interval' => substr( $subscription_phase->get_interval()->get_specification(), -2, 1 ),
			'item'     => array(
				'name'     => $subscription->description,
				'amount'   => $subscription_phase->get_amount()->format() * 100,
				'currency' => $payment_currency,
			),
			'notes'    => array(
				'knitpay_subscription_id'  => $subscription->get_id(),
				'knitpay_payment_id'       => $payment->id,
				'knitpay_subscription_key' => $subscription->get_key(),
			),
		);
		$razorpay_plan = $api->plan->create( $plan_data );

		$total_count = $this->get_max_count_for_period( $period, $subscription_phase->get_total_periods() );

		// TODO: Bug, total periods not updated in subscription
		// $subscription_phase->set_total_periods($total_count);

		$subscription_data     = array(
			'plan_id'         => $razorpay_plan->id,
			'total_count'     => $total_count,
			'customer_notify' => 1,
			'addons'          => array(),
			'notes'           => array(
				'knitpay_subscription_id'  => $subscription->get_id(),
				'knitpay_payment_id'       => $payment->id,
				'knitpay_subscription_key' => $subscription->get_key(),
			),
			'notify_info'     => array(
				'notify_phone' => $payment->get_billing_address()->get_phone(),
				'notify_email' => $customer->get_email(),
			),
		);
		$razorpay_subscription = $api->subs->create( $subscription_data );
		$razorpay_invoices     = $api->subs->get_subscription_invoices( $razorpay_subscription->id );

		// Save Subscription and Plan ID
		$subscription->set_meta( 'razorpay_subscription_id', $razorpay_subscription->id );
		$subscription->set_meta( 'razorpay_plan_id,', $razorpay_plan->id );

		$payment->set_transaction_id( $razorpay_invoices->items[0]->id );

		// Update gateway results in payment.
		$payment->set_action_url( $razorpay_subscription->short_url );
		$payment->recurring = true;

	}

	private function get_max_count_for_period( $period, $total_count ) {
		switch ( $period ) {
			case 'daily':
				return min( $total_count, 36500 );
			case 'weekly':
				return min( $total_count, 5200 );
			case 'monthly':
				return min( $total_count, 1200 );
			case 'yearly':
				return min( $total_count, 100 );
			default:
				return;
		}
	}

	private function handle_webhook( $api, $payment ) {
		$webhook_secret = $this->config->webhook_secret;

		$post = file_get_contents( 'php://input' );
		$data = json_decode( $post );

		if ( isset( $_SERVER['HTTP_X_RAZORPAY_SIGNATURE'] ) && isset( $webhook_secret ) ) {
			try {
				$api->utility->verifyWebhookSignature(
					$post,
					$_SERVER['HTTP_X_RAZORPAY_SIGNATURE'],
					$webhook_secret
				);
			} catch ( Errors\SignatureVerificationError $e ) {
				$log = array(
					'message' => $e->getMessage(),
					'data'    => $data,
					'event'   => 'razorpay.wc.signature.verify_failed',
				);

				error_log( json_encode( $log ) );
				return;
			}
		}

		$event_type = explode( '.', $data->event )[0];

		switch ( $event_type ) {
			case 'payment':
			case 'subscription':
				$razorpay_payment = $data->payload->payment->entity;

				$this->update_payment_status( $payment, $razorpay_payment );
				break;
		}

	}
}
